import torch
import torch.nn as nn
import torch.nn.functional as F


class Single_level_densenet(nn.Module):
    def __init__(self,filters, num_conv = 4):
        super(Single_level_densenet, self).__init__()
        self.num_conv = num_conv
        self.conv_list = nn.ModuleList()
        self.bn_list = nn.ModuleList()
        for i in range(self.num_conv):
            self.conv_list.append(nn.Conv2d(filters,filters,3, padding = 1))
            self.bn_list.append(nn.BatchNorm2d(filters))
            
    def forward(self,x):
        outs = []
        outs.append(x)
        for i in range(self.num_conv):
            temp_out = self.conv_list[i](outs[i])
            if i > 0:
                for j in range(i):
                    temp_out += outs[j]
            outs.append(F.relu(self.bn_list[i](temp_out)))
        out_final = outs[-1]
        del outs
        return out_final
    
class Down_sample(nn.Module):
    def __init__(self,kernel_size = 2, stride = 2):
        super(Down_sample, self).__init__()
        self.down_sample_layer = nn.MaxPool2d(kernel_size, stride)
    
    def forward(self,x):
        y = self.down_sample_layer(x)
        return y,x

class Upsample_n_Concat(nn.Module):
    def __init__(self,filters):
        super(Upsample_n_Concat, self).__init__()
        self.upsample_layer = nn.ConvTranspose2d(filters, filters, 4, padding = 1, stride = 2)
        self.conv = nn.Conv2d(2*filters,filters,3, padding = 1)
        self.bn = nn.BatchNorm2d(filters)
    
    def forward(self,x,y):
        x = self.upsample_layer(x)
        x = torch.cat([x,y],dim = 1)
        x = F.relu(self.bn(self.conv(x)))
        return x
    

class Dense_Unet(nn.Module):
    def __init__(self, in_chan, out_chan, filters, num_conv = 4):
        super(Dense_Unet, self).__init__()
        self.conv1 = nn.Conv2d(in_chan,filters,1)
        self.d1 = Single_level_densenet(filters,num_conv )
        self.down1 = Down_sample()
        self.d2 = Single_level_densenet(filters,num_conv )
        self.down2 = Down_sample()
        self.d3 = Single_level_densenet(filters,num_conv )
        self.down3 = Down_sample()
        self.d4 = Single_level_densenet(filters,num_conv )
        self.down4 = Down_sample()
        self.bottom = Single_level_densenet(filters,num_conv )
        self.up4 = Upsample_n_Concat(filters)
        self.u4 = Single_level_densenet(filters,num_conv )
        self.up3 = Upsample_n_Concat(filters)
        self.u3 = Single_level_densenet(filters,num_conv )
        self.up2 = Upsample_n_Concat(filters)
        self.u2 = Single_level_densenet(filters,num_conv )
        self.up1 = Upsample_n_Concat(filters)
        self.u1 = Single_level_densenet(filters,num_conv )
        self.outconv = nn.Conv2d(filters,out_chan, 1)
#         self.outconvp1 = nn.Conv2d(filters,out_chan, 1)
#         self.outconvm1 = nn.Conv2d(filters,out_chan, 1)
        
        
    def forward(self,x):
        bsz = x.shape[0]
        x = self.conv1(x)
        x,y1 = self.down1(self.d1(x))
        x,y2 = self.down1(self.d2(x))
        x,y3 = self.down1(self.d3(x))
        x,y4 = self.down1(self.d4(x))
        x = self.bottom(x)
        x = self.u4(self.up4(x,y4))
        x = self.u3(self.up3(x,y3))
        x = self.u2(self.up2(x,y2))
        x = self.u1(self.up1(x,y1))
        x1 = self.outconv(x)
#         xm1 = self.outconvm1(x)
#         xp1 = self.outconvp1(x)
        
        return x1


region_names_dict = {
        0: 'Unknown',
        1: 'Left-Lateral-Ventricle',
 2: 'Left-Inf-Lat-Vent',
 3: 'Left-Cerebellum-White-Matter',
 4: 'Left-Cerebellum-Cortex',
 5: 'Left-Thalamus-Proper',
 6: 'Left-Caudate',
 7: 'Left-Putamen',
 8: 'Left-Pallidum',
 9: '3rd-Ventricle',
 10: '4th-Ventricle',
 11: 'Brain-Stem',
 12: 'Left-Hippocampus',
 13: 'Left-Amygdala',
 14: 'CSF',
 15: 'Left-Accumbens-area',
 16: 'Left-VentralDC',
 17: 'Left-vessel',
 18: 'Left-choroid-plexus',
 19: 'Right-Cerebral-White-Matter',
 20: 'Right-Lateral-Ventricle',
 21: 'Right-Inf-Lat-Vent',
 22: 'Right-Cerebellum-White-Matter',
 23: 'Right-Cerebellum-Cortex',
 24: 'Right-Thalamus-Proper',
 25: 'Right-Caudate',
 26: 'Right-Putamen',
 27: 'Right-Pallidum',
 28: 'Right-Hippocampus',
 29: 'Right-Amygdala',
 30: 'Right-Accumbens-area',
 31: 'Right-VentralDC',
 32: 'Right-vessel',
 33: 'Right-choroid-plexus',
 34: 'Optic-Chiasm',
 35: 'CC_Posterior',
 36: 'CC_Mid_Posterior',
 37: 'CC_Central',
 38: 'CC_Mid_Anterior',
 39: 'CC_Anterior',
 40: 'ctx-lh-caudalanteriorcingulate',
 41: 'ctx-lh-caudalmiddlefrontal',
 42: 'ctx-lh-cuneus',
 43: 'ctx-lh-entorhinal',
 44: 'ctx-lh-fusiform',
 45: 'ctx-lh-inferiorparietal',
 46: 'ctx-lh-inferiortemporal',
 47: 'ctx-lh-isthmuscingulate',
 48: 'ctx-lh-lateraloccipital',
 49: 'ctx-lh-lateralorbitofrontal',
 50: 'ctx-lh-lingual',
 51: 'ctx-lh-medialorbitofrontal',
 52: 'ctx-lh-middletemporal',
 53: 'ctx-lh-parahippocampal',
 54: 'ctx-lh-paracentral',
 55: 'ctx-lh-parsopercularis',
 56: 'ctx-lh-parsorbitalis',
 57: 'ctx-lh-parstriangularis',
 58: 'ctx-lh-pericalcarine',
 59: 'ctx-lh-postcentral',
 60: 'ctx-lh-posteriorcingulate',
 61: 'ctx-lh-precentral',
 62: 'ctx-lh-precuneus',
 63: 'ctx-lh-rostralanteriorcingulate',
 64: 'ctx-lh-rostralmiddlefrontal',
 65: 'ctx-lh-superiorfrontal',
 66: 'ctx-lh-superiorparietal',
 67: 'ctx-lh-superiortemporal',
 68: 'ctx-lh-supramarginal',
 69: 'ctx-lh-transversetemporal',
 70: 'ctx-lh-insula',
 71: 'ctx-rh-caudalanteriorcingulate',
 72: 'ctx-rh-caudalmiddlefrontal',
 73: 'ctx-rh-cuneus',
 74: 'ctx-rh-entorhinal',
 75: 'ctx-rh-fusiform',
 76: 'ctx-rh-inferiorparietal',
 77: 'ctx-rh-inferiortemporal',
 78: 'ctx-rh-isthmuscingulate',
 79: 'ctx-rh-lateraloccipital',
 80: 'ctx-rh-lateralorbitofrontal',
 81: 'ctx-rh-lingual',
 82: 'ctx-rh-medialorbitofrontal',
 83: 'ctx-rh-middletemporal',
 84: 'ctx-rh-parahippocampal',
 85: 'ctx-rh-paracentral',
 86: 'ctx-rh-parsopercularis',
 87: 'ctx-rh-parsorbitalis',
 88: 'ctx-rh-parstriangularis',
 89: 'ctx-rh-pericalcarine',
 90: 'ctx-rh-postcentral',
 91: 'ctx-rh-posteriorcingulate',
 92: 'ctx-rh-precentral',
 93: 'ctx-rh-precuneus',
 94: 'ctx-rh-rostralanteriorcingulate',
 95: 'ctx-rh-rostralmiddlefrontal',
 96: 'ctx-rh-superiorfrontal',
 97: 'ctx-rh-superiorparietal',
 98: 'ctx-rh-superiortemporal',
 99: 'ctx-rh-supramarginal',
 100: 'ctx-rh-transversetemporal',
 101: 'ctx-rh-insula',
 102: 'Left-Cerebral-White-Matter'
      }



import pickle
import numpy as np
import torch
import nibabel
import torch
import torch.nn as nn
import torch.nn.functional as F


def pickling(file,path):
    pickle.dump(file,open(path,'wb'))
def unpickling(path):
    file_return=pickle.load(open(path,'rb'))
    return file_return

def dice_score(pred,gt, ep= 1e-4):
    sh1,sh2,sh2,C = pred.shape
#     print(pred.shape)
#     print(gt.shape)
    score_list = []
    for i in range(C-1):
        num = 2*(np.sum(pred[:,:,:,i]*gt[:,:,:,i])) + ep
        denom = np.sum(pred[:,:,:,i] + gt[:,:,:,i]) + ep
        score = num/denom
        score_list.append(score)
    count = np.sum(np.transpose(gt,axes = (1,0,2,3)).reshape(C,-1),axis = 1)
    return score_list,count

def load_data(path, mgz = False):
    if mgz:
        t1_img_nii = nibabel.MGHImage.from_filename(path)
    else:
        t1_img_nii = nibabel.load(path)
    affine_map = t1_img_nii.affine
    t1_img, orientation = orient_correctly(t1_img_nii)
    sh1, sh2, sh3 = t1_img.shape
    
    
    dir1_pad = (256 - sh1)//2
    if dir1_pad < 0:
        t1_img = t1_img[-dir1_pad:dir1_pad+ sh1%2,:,:]
    else:
        t1_img = np.pad(t1_img,((dir1_pad,dir1_pad + sh1%2),(0,0),(0,0)),mode = 'constant',constant_values = 0.0)
        
        
    dir2_pad = (256 - sh2)//2
    if dir2_pad < 0:
        t1_img = t1_img[:,-dir2_pad:dir2_pad+ sh2%2,:]
    else:
        t1_img = np.pad(t1_img,((0,0),(dir2_pad,dir2_pad + sh2%2),(0,0)),mode = 'constant',constant_values = 0.0)
        
     
    dir3_pad = (256 - sh3)//2
    if dir3_pad < 0:
        t1_img = t1_img[:,:,-dir3_pad:dir3_pad+ sh3%2]
    else:
        t1_img = np.pad(t1_img,((0,0),(0,0),(dir3_pad,dir3_pad + sh3%2)),mode = 'constant',constant_values = 0.0)
        
#     t1_img = np.pad(t1_img,((dir1_pad,dir1_pad + sh1%2),(dir2_pad,dir2_pad + sh2%2),(dir3_pad,dir3_pad + sh3%2)), \
#                     mode = 'constant',constant_values = 0.0)
    return t1_img,orientation, dir1_pad,sh1,dir2_pad,sh2, dir3_pad, sh3, affine_map

def orient_correctly(img_nii):
    orientation = nibabel.io_orientation(img_nii.affine)
    try:
        img_new = nibabel.as_closest_canonical(img_nii, True).get_data().astype(float)
    except:
        img_new = nibabel.as_closest_canonical(img_nii, False).get_data().astype(float)
    img_trans = np.transpose(img_new, (0,2,1))
    img_flip = np.flip(img_trans,0)
    img_flip = np.flip(img_flip,1)
    return img_flip, orientation

def orient_to_ras(image):
    img_flip = np.flip(image,0)
    img_flip = np.flip(img_flip,1)
    if image.ndim > 3:
        img_trans = np.transpose(img_flip,(0,2,1,3))
    else:
        img_trans = np.transpose(img_flip,(0,2,1))
    return img_trans

def back_to_original_4_pred(image,orientation, dir1_pad,sh1, dir2_pad,sh2, dir3_pad,sh3):
    if dir1_pad < 0:
        image = np.pad(image,((-dir1_pad,-dir1_pad- sh1%2),(0,0),(0,0)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[dir1_pad:dir1_pad+sh1,:,:]
        
    if dir2_pad < 0:
        image = np.pad(image,((0,0),(-dir2_pad,-dir2_pad - sh2%2),(0,0)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[:,dir2_pad:dir2_pad+sh2,:]
        
    if dir3_pad < 0:
        image = np.pad(image,((0,0),(0,0),(-dir3_pad,-dir3_pad- sh3%2)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[:,:,dir3_pad:dir3_pad+sh3]
    
#     img_unpadded = image[dir1_pad:dir1_pad+sh1,dir2_pad:dir2_pad+sh2,dir3_pad:dir3_pad+sh3]
    img_ras = orient_to_ras(image)
    img_orig_orient = np.transpose(img_ras, orientation[:,0].astype(int))
    for k,i in enumerate(orientation[:,1]):
        if i == -1.0:
            img_orig_orient = np.flip(img_orig_orient,k)
    
    return img_orig_orient

def back_to_original_4_prob(image,orientation, dir1_pad,sh1, dir2_pad,sh2, dir3_pad,sh3):
    if dir1_pad < 0:
        image = np.pad(image,((-dir1_pad,-dir1_pad- sh1%2),(0,0),(0,0),(0,0)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[dir1_pad:dir1_pad+sh1,:,:,:]
        
    if dir2_pad < 0:
        image = np.pad(image,((0,0),(-dir2_pad,-dir2_pad - sh2%2),(0,0),(0,0)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[:,dir2_pad:dir2_pad+sh2,:,:]
        
    if dir3_pad < 0:
        image = np.pad(image,((0,0),(0,0),(-dir3_pad,-dir3_pad- sh3%2),(0,0)),mode = 'constant',constant_values = 0.0)
    else:
        image = image[:,:,dir3_pad:dir3_pad+sh3,:]
#     img_unpadded = image[dir1_pad:dir1_pad+sh1,dir2_pad:dir2_pad+sh2,dir3_pad:dir3_pad+sh3,:]
    img_ras = orient_to_ras(image)
    transpose_axis = orientation[:,0].astype(int)
    transpose_axis = np.append(transpose_axis, 3)
    img_orig_orient = np.transpose(img_ras, transpose_axis)
    for k,i in enumerate(orientation[:,1]):
        if i == -1.0:
            img_orig_orient = np.flip(img_orig_orient,k)
    
    return img_orig_orient

def create_one_hot_seg(image,num_seg):
    p = F.softmax(image,dim = 1)
    p_maxim = (torch.max(p, dim=1)[1]).cpu().data.numpy()
    img = []
    for seg in range(num_seg):
        masked = np.expand_dims((p_maxim==seg).astype(float),axis = 1)
        img.append(masked)
    return np.concatenate(img,axis = 1)



#-*- conding: utf-8 -*-

"""
DARTS
"""

__author__ = "Razavian Lab"
__email__ = "tms598@nyu.edu"
__copyright__ = "Copyright (C) 2019, Razavian Lab"


from DARTS.models.dense_unet_model import Single_level_densenet, Down_sample, Upsample_n_Concat, Dense_Unet
from DARTS.utils import load_data, back_to_original_4_pred

from DARTS.darts import Segmentation
from DARTS._version import __version__



import os
import torch
import numpy as np
import pandas as pd
import nibabel
from sys import exit
from DARTS.models.dense_unet_model import Dense_Unet
from DARTS.utils import load_data, back_to_original_4_pred
from DARTS.region_names import region_names_dict

class Segmentation():
    def __init__(self, model_wts_path='./dense_unet_back2front_finetuned.pth'):
        """
        Initializes the segmentation object (GPU required)
        Parameters
        ----------
        model_wts_path : Path to pre-trained model (default './dense_unet_back2front_finetuned.pth')
        """
        super(Segmentation, self).__init__()

        if torch.cuda.is_available():
            self.device = 'cuda'
            self.model = Dense_Unet(in_chan = 1, out_chan = 113, filters = 256, num_conv = 4)
            self.model.load_state_dict(torch.load(model_wts_path))
            self.model = self.model.to(self.device)
            self.model.eval()
            self.input_image = None
        else:
            exit("GPU required")

    def read_image(self, input_image_path):
        """
        Normalizes specified input and output as np.array
        Parameters
        ----------
        input_image_path : Path to input image (can be .mgh, .mgz, .nii or .nii.gz)

        Returns
        -------
        Normalized input MRI as np.array
        """

        if ".mg" in str.lower(input_image_path):
            is_mgz = True
        else:
            is_mgz = False

        input_image, orient, pad_sh1, sh1, pad_sh2, sh2, pad_sh3, sh3, affine_map = load_data(input_image_path, is_mgz)
        input_image[np.isnan(input_image)] = 0
        input_image = np.clip(input_image, a_min=0.0, a_max=np.max(input_image))

        remove_idx_list = [34, 35, 42, 43, 73, 74, 77, 78, 108, 109]
        self.keep_idx_list = [i for i in range(113) if i not in remove_idx_list]

        self.sh2 = input_image.shape[2]
        input_image = torch.from_numpy(input_image).to(self.device).float()
        for i in range(self.sh2):
            max_value = torch.max(input_image[:, :, i])
            min_value = torch.min(input_image[:, :, i])
            input_image[:, :, i] = (input_image[:, :, i] - min_value) / (max_value - min_value + 1e-4)

        self.input_image = input_image
        self.orient, self.pad_sh1, self.sh1, self.pad_sh2, self.sh2, self.pad_sh3, self.sh3, self.affine_map = orient, pad_sh1, sh1, \
                                                                                              pad_sh2, sh2, pad_sh3, sh3, affine_map

        return self.input_image

    def predict(self, input_image_path, output_dir='./'):
        """
        Segments the input image
        Parameters
        ----------
        input_image_path : Path to input image (can be .mgh, .mgz, .nii or .nii.gz)
        output_dir : Path to save segmentation output and volumes (default: './')

        Returns
        -------
        Segmentation mask (np.array)
        """
        _ = self.read_image(input_image_path)
        print("Segmenting image")
        sh2 = self.input_image.shape[2]
        keep_idx_list = self.keep_idx_list
        pred_arg_max = [0] * sh2  # stores the prediction
        for i in range(sh2):
            out = self.model(self.input_image[:, :, i].unsqueeze(0).unsqueeze(1))
            out = out[:, keep_idx_list, :, :]
            pred_arg_max[i] = torch.argmax(out, dim=1).data.cpu().numpy()
        pred_arg_max = np.moveaxis(np.concatenate(pred_arg_max, axis=0), 0, -1)
        pred_orig = back_to_original_4_pred(pred_arg_max, self.orient, self.pad_sh1, self.sh1, self.pad_sh2, \
                                            self.sh2, self.pad_sh3, self.sh3)
        #change labels
        pred_orig[pred_orig == 102] = 103
        pred_orig[pred_orig == 0] = 102
        pred_orig[pred_orig == 103] = 0

        print("Segmentation complete, saving predictions")

        self.save_output_one(pred_orig, input_image_path, output_dir)
        return pred_orig

    def save_output_one(self, pred_orig, input_image_path, output_dir='./'):
        """
        Saves the input image and volumes
        Parameters
        ----------
        pred_orig : Segmentation mask (np.array)
        input_image_path : Path to input image (can be .mgh, .mgz, .nii or .nii.gz)
        output_dir : Path to save segmentation output and volumes (default: './')
        """

        file_name = input_image_path.split('.')[0].split('/')[-1]
        if ".mg" in str.lower(input_image_path):
            pred_orig_nib = nibabel.freesurfer.mghformat.MGHImage(pred_orig.astype(np.float32), None)
            nibabel.save(pred_orig_nib, os.path.join(output_dir, file_name+'_seg.mgz'))
        else:
            pred_orig_nib = nibabel.Nifti1Image(pred_orig, None)
            nibabel.save(pred_orig_nib, os.path.join(output_dir, file_name+'_seg.nii.gz'))

        if ".mg" in str.lower(input_image_path):
            seg = nibabel.freesurfer.MGHImage.from_filename(os.path.join(output_dir,file_name+'_seg.mgz'))
            if np.sum(seg.affine[:,:3] != self.affine_map[:,:3]) != 0:
                pred_orig_nib = nibabel.freesurfer.mghformat.MGHImage(seg.get_data().astype(np.float32), self.affine_map)
                nibabel.save(pred_orig_nib, os.path.join(output_dir, file_name+'_seg.mgz'))
        else:
            seg = nibabel.load(os.path.join(output_dir, file_name+'_seg.nii.gz'))
            if np.sum(seg.affine[:, :3] != self.affine_map[:,:3]) != 0:
                pred_orig_nib = nibabel.Nifti1Image(seg.get_data(), self.affine_map)
                nibabel.save(pred_orig_nib, os.path.join(output_dir, file_name+'_seg.nii.gz'))

        # make volumes dictionary
        unique, counts = np.unique(pred_orig_nib.get_data(), return_counts=True)
        volumes = counts * (pred_orig_nib.header.get_zooms()[0] * pred_orig_nib.header.get_zooms()[1] * pred_orig_nib.header.get_zooms()[2])
        volumes_dict= dict(zip(unique.astype(int), volumes))
        # save to csv
        df = pd.DataFrame.from_dict(volumes_dict, orient = "index", columns=["Volume (voxels)"])
        df.index.name = "Label Number"
        dict_2 = {}
        for i in unique:
            dict_2[i] = region_names_dict[i]
        df.insert(0, "Region", dict_2.values())
        df.to_csv(os.path.join(output_dir, file_name+"_volumes.csv"))

        print("Segmented image and volumes saved at", output_dir)



from setuptools import setup, find_packages
from os.path import join as pjoin
from DARTS._version import __version__

requirements= open('requirements.txt').read().split()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='DARTS',
    version=__version__,
    description='DenseUnet-based Automatic Rapid brain Segmentation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='NYUMedML',
    author_email='tms598@nyu.edu',
    url='https://github.com/NYUMedML/DARTS',
    license='GNU General Public License v3.0',
    install_requires=requirements,
    packages=find_packages(exclude=('archive','plots')),
    package_data={'models': [pjoin('models','*')]},
    python_requires='>=3')
